; ------------------------------------------------------------
;; William R. Bevier and Michael K. Smith.
;; Copyright 1994, Computational Logic, Inc. All Rights Reserved
;; ------------------------------------------------------------

(require "general")
(require "nodes")
(provide "arcs")

;; ------------------------------------------------------------
;; We use the term `arc' to mean transitions between nodes. 
;; Like nodes, arcs are named entities.

;; CHANGES. 
;; 1. We have made arcs 1-to-1 rather than many-to-many.
;; 2. Arcs cannot dive into AND nodes (which is pretty much redundant).
;;
;; Associated with each arc is 
;;     from, a source node, 
;;     to,   a target node,
;;     [ trigger ], a trigger
;;     [ action ], an action
;;     scope, a node name which is computed.  The highest node
;;      shared by from and to.
;; 
;; ------------------------------------------------------------

(eval-when
 (compile load eval)

 (defstruct arc name from to (trigger t) (action nil) scope)

 (defmacro update-arc (x &rest rest)
   `(make-arc ,@(update-fields 'arc
			       x
			       '(name from trigger action to scope)
			       rest)))
  )

;; ARCP
;; The sources (from) and targets (to) of an arc are sets of symbols
;; that are also the names of nodes. The targets are non
;; empty lists. NIL is not an allowed node name.

(defun arcp (x nodes)
  (let ((name (arc-name x))
        (from (arc-from x))
	(to   (arc-to x))
	;;(trigger (arc-trigger x))
	;;(action (arc-action x))
	)
    (and (static-check (arc-p x)
		      "~%ARCP: ~a not an arc struct." x)
	 (static-check (symbolp from)
		      "~%ARCP: From ~a of arc ~a must be a symbol." from name)
	 (static-check (symbolp to)
		      "~%ARCP: To ~a of arc ~a must be a symbol." to name)
	 (let ((names (node-names nodes)))
	   (and (static-check (member from names)
			     "~%ARCP: From ~a not a node name in arc ~a." from name)
		(static-check (member to names)
			     "~%ARCP: To ~a not a node name in arc ~a." to name)) )
	 )))

(defun full-arcp (x nodes)
  (let ((name (arc-name x))
        (scope (arc-scope x))
	(to   (arc-to x)))
    (static-check (to-node-and-free-up-to-scope to scope nodes)
		 "~%ARCP: ~a penetrates an AND on its way to ~a." name to)))

(defun to-node-and-free-up-to-scope (nm scope nodes)
  (let* ((n (find-node nm nodes))
	 (np (node-parent n)))
    (or (equal nm scope)
	(null np)
	(and (not (equal (ntype np nodes) 'and))
	     (to-node-and-free-up-to-scope np scope nodes)))))

;; An arc list must be well-formed with respect to a node list.

;; Forall arcs, 
;;   ARCP must be true before (COMPUTE-STATIC-ARC-INFORMATION arcs), 
;;   FULL-ARCP after.


;; ARC-LISTP assumes that NODES is a list of nodes
;; to which the function COMPUTE-STATIC-NODE-INFORMATION has
;; been applied.

(defun arc-listp (arcs nodes)
  (iterate
   for a in arcs
   always (arcp a nodes)))

(defun full-arc-listp (arcs nodes)
  (iterate
   for a in arcs
   always (full-arcp a nodes)))

(defun actions-and-triggers-ok (arcs prim-events names vars)
  (iterate
   for a in arcs
   always (let ((aname (arc-name a))
		(trigger  (arc-trigger a))
		(action (arc-action a)))
	    (and (static-check (or (null trigger) (eventp trigger prim-events names vars))
			      "~%ACTIONS&TRIGGERS: Trigger ~a of arc ~a not a condition." trigger aname)
		 (static-check (or (null action) (actionp action prim-events names vars))
			      "~%ACTIONS&TRIGGERS: Action ~a of arc ~a not an action." action aname)))))
	

;; From page 22, Semantics of Statecharts.
;; The scope of an arc is the lowest OR node in the hierarchy
;; of nodes which is the proper common ancestor of all
;; the sources and targets of the arc.

(defun scope-of-arc (arc nodes)
  (deepest-containing-OR-node
   (deepest-containing-node (list (arc-from arc) (arc-to arc))
			    nodes)
   nodes))

;; ------------------------------------------------------------
;; Accessor fns for lists of arcs
;; ------------------------------------------------------------

(defun find-arc (nm arcs)
  (iterate for a in arcs
    when (eq nm (arc-name a))
    return a))

(defun afrom (nm arcs)
  (arc-from (find-arc nm arcs)))

(defun ato (nm arcs)
  (arc-to (find-arc nm arcs)))

(defun ascope (nm arcs)
  (arc-scope (find-arc nm arcs)))

(defun atrigger (nm arcs)
  (arc-trigger (find-arc nm arcs)))

(defun aaction (nm arcs)
  (arc-action (find-arc nm arcs)))
