
;; ------------------------------------------------------------
;; William R. Bevier and Michael K. Smith.
;; Copyright 1994, 1996 Computational Logic, Inc. All Rights Reserved
;; ------------------------------------------------------------

(require "iterate")
(require "general")
(require "system")
(provide "conditions-events-actions")

;; ------------------------------------------------------------
;; CHANGES for VERSION 3.
;
;; 1. Allow defined conditions and actions as Lisp functions.
;; 2. Delete local vars, as Lisp provides them.
;
;; Open issues:
;;    
;;    A number of action forms are stubbed.
;
;; ------------------------------------------------------------

;; Define recognizers for Conditions and Actions.
;; We define evaluators for these kinds of expressions, as well as
;; for Exprs, which are just expressions on state variables.
;
;; The original recognizer definitions were taken from the appendix of 
;; The Languages of Statemate.
;
;; In the following:
;
;; prim-events is a list of legal primitive events
;; states      is a list of node names
;; vars        is a list of legal variables names

;; Unfinished:
;; We're ignoring structured variables in the expression language 
;;   (e.g., lists, records, groups). 

(defun conditionp (form states vars)
  (cond ((symbolp form) (assoc form vars))
	((consp form)
	 (case (car form)
	   (in                     (consp (member (cadr form) states)))
	   ((eq neq < >)           (equal (length form) 3))
	   (not                    (conditionp (cadr form) states vars))
	   ((and or)               (iterate for x in (cdr form)
					    always (conditionp x states vars)))
	   (otherwise
	    (or (cboundp (car form))
		(fboundp (car form))))))
	(t nil)))

(defun eventp (form prim-evs states vars)
  (cond ((symbolp form) (member form prim-evs))
	((consp form)
	 (case (car form)
	   ((entered exited)          (consp (member (cadr form) states)))
	   ((changed read written)    (assoc (cadr form) vars))
	   ((true false)              (conditionp (cadr form) states vars))
	   (timeout                   (numberp (cadr form))) ;; ????
	   (cond     (and (or (null (cadr form))
			      (eventp (cadr form) prim-evs states vars))
			  (conditionp (caddr form) states vars)))
	   (not      (eventp (cadr form) prim-evs states vars))
	   ((and or) (iterate for x in (cdr form)
		       always (eventp x prim-evs states vars)))
	   (notify   (patternp (cadr form)))
	   (otherwise (or (eboundp (car form))
			  (cboundp (car form))
			  (fboundp (car form))))))
	(t nil)))

(defun actionp (form prim-evs states vars)
  (cond ((symbolp form) (consp (member form prim-evs)))
	((consp form)
	 (case (car form)
	   ((make-true make-false)     (assoc (cadr form) vars))
	   (assign                     (and (assoc (cadr form) vars) (equal (length form) 3)))
	   ((read-data write-data)     (assoc (cadr form) vars))
	   ;; ((history-clear deep-clear) (assoc (cadr form) states))
	   (schedule   (and (equal (length form) 3)
			    (numberp (cadr form))
			    (actionp (caddr form) prim-evs states vars)))
	   (par        (iterate for x in (cdr form) 
				always (actionp x prim-evs states vars)))
	   (if         (and (conditionp (cadr form) states vars)
			    (actionp (caddr form) prim-evs states vars)
			    (or (< (length form) 4)
				(actionp (cadddr form) prim-evs states vars))))
	   (when       (and (eventp (cadr form) prim-evs states vars)
			    (actionp (caddr form) prim-evs states vars)
			    (or (< (length form) 4)
				(actionp (cadddr form) prim-evs states vars))))
	   (t (or (aboundp (car form))
		  (fboundp (car form))))))
	(t nil)))

(defun patternp (form)
  (cond ((equal form nil) t)
	((equal form '*) t)
	((symbolp form) t)
	((not (consp form)) nil)
	((equal (car form) 'one-of) (all-patternp (cdr form)))
	((equal (car form) '-) (patternp (cadr form)))
	((equal (car form) '@) (symbolp (cadr form)))
	((equal (car form) '?) (symbolp (cadr form)))
	((symbolp (car form)) (partlistp (cdr form)))
	((and (consp (car form)) (eq (car (car form)) '-))
	 (partlistp (cdr form)))
	(t nil)))

(defun all-patternp (form)
  (if (null form) t 
      (if (and (consp form) (patternp (car form)))
	  (all-patternp (cdr form))
	  nil)))

(defun partlistp (l)
  (cond ((null l) t)
	((partp (car l)) (partlistp (cdr l)))
	(t nil)))

(defun partp (l)
  (cond ((null l) t)
	((atom l) t)
	((partp (car l)) (partp (cdr l)))
	(t nil)))

;; ------------------------------------------------------------	       
;; Evaluators for Conditions, Events, Actions and Expressions.
;; ------------------------------------------------------------

;; form          is a form to be evaluated
;; evs           is a list of event forms, considered to be current
;; config        is a set of state names that is a legal configuration
;; vmap          is an alist mapping global variables to values

(defvar *nodes* nil "Set by initial-system-step")

(defvar *time* 0)

(defun eval-expr (form config vmap)
  "Evals conditions as well as expressons in actions."
  (cond ((symbolp form) (cdr (assoc form vmap)))
	((not (consp form)) form)
	((equal (car form) 'in) (member (cadr form) config))
	((cboundp (car form))
	 (eval-expr (lookup (car form) *cond-defs*) config vmap))
	(t 
	 (case (car form)
	   (busy (node-busy (find-node (cadr form) *nodes*)))
	   (idle (compute-node-idle (find-node (cadr form) *nodes*) *time*))
	   (eq (equal (eval-expr (nth 1 form) config vmap) 
		      (eval-expr (nth 2 form) config vmap)))
	   (neq (not (equal (eval-expr (nth 1 form) config vmap) 
			    (eval-expr (nth 2 form) config vmap))))
	   (and (and (eval-expr (nth 1 form) config vmap) 
		     (eval-expr (nth 2 form) config vmap)))
	   (not (not (eval-expr (cadr form) config vmap)))
	   (or  (or  (eval-expr (nth 1 form) config vmap) 
		     (eval-expr (nth 2 form) config vmap)))
	   ;; Handled above (in may be CL func)
	   ;; (in (member (cadr form) conf))
	   (otherwise 
	    (if (fboundp (car form))
		(apply (car form) (iterate for arg in (cdr form)
					   collect (eval-expr arg config vmap)))
		(error "EVAL-EXPR: Can't handle ~a." form)))))))

(defun compute-node-idle (n time)
  ;; time is ticks.
  (let ((idle (node-idle n)))
    (if (and (numberp idle) (> idle 0))
	(- time idle)
	0)))

;; Implicit guard: (eventp form ...)
;; Returns boolean.

(defvar *bindings* nil "Used to bind (? vars) in events for later use by actions.")

(defun eval-event (form events time conf vmap)
  (let (*bindings*)
    (if (eval-event2 form events time conf vmap)
	(cons 't *bindings*)
	nil)))

(defun eval-event2 (form events time conf vmap)
  (cond ((symbolp form) (consp (member form events :test #'equal)))
	((consp form)
	 (cond ((eboundp (car form))
		(eval-event2 (lookup (car form) *event-defs*) events time conf))
	       ((cboundp (car form))
		(eval-expr (lookup (car form) *cond-defs*) conf vmap))
	       (t (case (car form)
		    ((entered exited changed true false read written)
		     ;; started st stopped sp
		     (consp (member form events :test #'equal)))
		    (cond (if (null (cadr form))
			      (eval-expr (caddr form) conf vmap)
			      (and (eval-event2 (cadr form) events time conf vmap)
				   (or (null (caddr form))
				       (eval-expr (caddr form) conf vmap)))))

		    ;; (timeout n e) means n time units have passed
		    ;; since e last occured.
		    (timeout (check-timeout (cadr form) (caddr form) time))

		    (and (and (eval-event2 (nth 1 form) events time conf vmap) 
			      (eval-event2 (nth 2 form) events time conf vmap)))
		    (not (not (eval-event2 (cadr form)  events time conf vmap)))
		    (or  (or  (eval-event2 (nth 1 form) events time conf vmap) 
			      (eval-event2 (nth 2 form) events time conf vmap)))
		    (in (member (cadr form) conf))

		    (notify (eval-pattern (cadr form) events time conf vmap))

		    (t nil)))))
	(t (error "EVAL-EVENT2: Can't handle ~a." form))))

(defvar *event-history* nil)

(defun check-timeout (etime e time)
  ;; NEEDS OPTIMIZATION.
  ;; Note that TIMEs are TICK TIME, not REAL TIME.
  (let ((previous (find-event e *event-history*)))
    (cond ((null previous) nil)
	  ((>= (- time (cdr previous)) etime) t)
	  (t nil))))

(defun find-event (e history)
  ;; history of form ((event . time)*)
  ;; E.g. (((NOTIFY (HELP (TEXT . "look") (PLAYER . |#14792|) ..))  . 30)
  ;;       ((READ TICK) . 29) ((CHANGED TICK) . 29) ((WRITTEN TICK) . 29)
  ;;       ((NOTIFY (HELP (TEXT . "look") (PLAYER . |#14792|) ..))  . 29))
  (cond ((null history) nil)
	((eq e '*)   (find-event-atom 'notify history))
	((symbolp e) (find-event-atom e history))
	(t ;(find-event-prefix e history)
	 nil)))

(defun find-event-atom (e history)
  (cond ((null history) nil)
	((eq e (car (car (car history)))) (car history))
	(t (find-event-atom e (cdr history)))))

(defun eval-pattern (form events time conf vmap)
  (eval-pattern2 form (collect-notifies events) time conf vmap))

(defun eval-pattern2 (form notifys time conf vmap)
  ;; Notifys is a list of events, each of which of of the form (type . args)
  (cond ((equal form nil) (not notifys))
	((equal form '*) notifys)
	((null notifys) nil)
	((symbolp form) (subset-notifies form notifys))
	((not (consp form)) nil)
	((equal (car form) '-) (not (eval-pattern2 (cadr form) notifys time conf vmap))) 
	((equal (car form) 'one-of)
	 (iterate for pat in (cdr form)
		  when (eval-pattern2 pat notifys time conf vmap)
		  return t)) 
	((symbolp (car form))
	 (match-parts (subset-notifies (car form) notifys) (cdr form) vmap))
	((and (consp (car form)) (eq (car (car form)) '-))
	 (match-parts (subset-notifies (car form) notifys) (cdr form) vmap))
	(t nil)))

(defun collect-notifies (l)
  (iterate for x in l
	   when (and (consp x) (equal (car x) 'notify))
	   collect (cadr x)))

(defun subset-notifies (sym l)
  ;; Sym is a symbol or (- symbol)
  (iterate for x in l
	   when (and (consp x)
		     (cond ((symbolp sym)
			    (equal (car x) sym))
			   ((and (consp sym) (eq (car sym) '-))
			    (not (equal (car x) (car (cdr sym)))))))
	   collect x))

(defun match-parts (posslist pattern vmap) 
  ;; pattern = (cdr '(symbol [ part* ]))
  (cond ((not (consp posslist)) nil)
	;; Car has already been matched.  Pattern must match cdr.
	((match-part-cdr (cdr (car posslist)) pattern vmap) t)
	(t (match-parts (cdr posslist) pattern vmap))))

(defun match-part-cdr (poss pattern vmap) 
  ;; pattern = (cdr '(symbol [ part* ]))
  (cond ((null pattern) t)	
	((eq pattern '*) t)
	((equal pattern poss) t)
	((not (consp pattern)) nil)
	((not (consp poss)) nil)
	((match-part (car poss) (car pattern) vmap)
	 (match-part-cdr (cdr poss) (cdr pattern) vmap))
	(t nil)))

(defun match-part (poss part vmap) 
  ;; Part and poss may be dotted pairs.
  (cond ((equal part '*) t)
	((atom part)
	 (if (consp poss)
	     (equal part (car poss))
	     (equal part poss)))
	((not (consp part)) nil)
	((eq (car part) '?)
	 (setq *bindings* (cons (cons (cadr part) poss) *bindings*))
	 t)
	((eq (car part) '@)
	 (let ((value (assoc (cadr part) vmap)))
	   (if (and value (or (not (consp value)) (not (eq (car value) '@))))
	       (match-part poss (cdr value) vmap)
	       nil)))
	((eq (car part) (car poss))
	 (match-part (cdr poss) (cdr part) vmap))
	(t nil)))


;; Returns a multiple value (evs vmap)
;; Each of these values represents changes to the original parameter made 
;; during evaluation.
;;
;; Implicit guard: (actionp form ...)

(defun eval-action (form system)
  ;; The global *bindings* may be bound based on (? symbol)
  ;; matches to the event that triggered this
  ;; action.  
  (let* ((config (system-configuration system))
	 (bindings *bindings*)
	 (vmap (bind-all *bindings* (system-alist system)))
	 (*bindings* nil))
    (setf (system-alist system) vmap)
    (cond ((symbolp form);; 
	   ;; Assertion of a primitive event.
	   (values (list (list 0 form)) bindings))
	  ((consp form)
	   (if (aboundp (car form))
	       (eval-action (lookup (car form) *action-defs*) system)
	       (case (car form)
		 ((make-true);; the arg is a primitive condition, a global
		  (let ((pair (assoc (cadr form) vmap)))
		    (cond (pair (let ((b (cdr pair)))
				  (if b 
				      (values nil bindings)
				      (values (list (cons 0 `(tr ,(cadr form))))
					      (bind (cadr form) t bindings)))))
			  (t (error "EVAL-ACTION: Condition ~a not known." form)))))
		 ((make-false);; the arg is a primitive condition, a global
		  (let ((pair (assoc (cadr form) vmap)))
		    (cond (pair (let ((b (cdr pair)))
				  (if b (values (list (cons 0 `(fs ,(cadr form))))
						(bind (cadr form) nil bindings))
				      (values nil bindings))))
			  (t (error "EVAL-ACTION: Condition ~a not known." form)))))
		 (assign (let* ((lhs (cadr form))
				(rhs (caddr form))
				(oldval (eval-expr lhs config vmap))
				(newval (eval-expr rhs config vmap)))
			   (cond ((symbolp lhs);; a global variable
				  (values
				   `(( 0 . (written ,lhs))
				     ,@(if (equal oldval newval) nil `((0 . (changed ,lhs))))
				     ;; We think this should be included, but haven't confirmed this.
				     ,@(iterate for var in (all-vars rhs)
						collect `(0 . (read ,var))))
				   (bind lhs newval bindings)))
				 (t (error "EVAL-ACTION: Bad assign statement lhs ~a." form)))))
		 ;; ((start stop suspend resume) (error "EVAL-ACTION: can't handle ~a." form))
		 (schedule (values (list (cdr form)) bindings))
		 (reschedule (values (list (list '* (cadr form)) (cdr form)) bindings))
		 (par (eval-par-args (cdr form) system))
		 (do  (do-actions (cdr form) system))
		 (if  (if (eval-expr (cadr form) config vmap)
			  (eval-action (caddr form) system)
			  (if (< (length form) 4)
			      (values nil bindings)
			      (eval-action (cadddr form) system))))
		 (t (if (fboundp (car form))
			(progn
			  (apply (car form) (iterate for arg in (cdr form)
						   collect (eval-expr arg config vmap)))
			  (values nil bindings))
			(error "EVAL-ACTION: unrecognized form ~a" form))))))
	  (t (error "EVAL-ACTION: unrecognized form ~a" form)))))

(defun eval-par-args (forms system)
  ;; FORMs is a list of forms, the args to a PAR action.
  (cond ((null forms) (values nil nil))
	(t (multiple-value-bind (a b)
	       (eval-action (car forms) system)
	     (multiple-value-bind (x y)
		 (eval-par-args (cdr forms) system)
	       (values (union a x :test #'equal);; merge events
		       (merge-alists b y t);; var alist
		       ))))))

(defun do-actions (forms system)
  ;; FORMs is a list of forms, the args to a PAR action.
  (cond ((null forms) (values nil nil))
	(t (multiple-value-bind (a b)
	       (eval-action (car forms) system)
	     (multiple-value-bind (x y)
		 (do-actions (cdr forms) system)
	       (values (union a x :test #'equal);; merge events
		       (merge-alists b y t);; var alist
		       ))))))
	     
(defun all-vars (form)
  (cond ((consp form)
	 (cond ((eq (car form) 'quote) nil)
	       ((eq (car form) 'lookup) nil)
	       (t (iterate for arg in (cdr form)
		    append (all-vars arg)))))
	((symbolp form)
	 (cond ((eq form t) nil)
	       ((eq form nil) nil)
	       (t (list form))))
	(t nil)))


