
;; ------------------------------------------------------------
;; Michael K. Smith.
;; Copyright 1996 Computational Logic, Inc. All Rights Reserved
;; ------------------------------------------------------------

(require "general")
(require "conditions-events-actions")
(require "system")
(require "nodes")
(require "arcs")
(require "step")

(provide "interactive")

(defvar *eval-mode* 'interactive)
(defvar *system* nil)

;; ----------------------------------------------------------------------
;; Interactive step

(defun irun (system)
  (setq *system* (initial-system-step system))
  (irun-system *system*))

(defun irun-system (system)
  (let ((live t) command result)
    (format t "~%+>")
    (iterate while (and live (not (member (setq command (read)) '(exit quit bye halt))))
	     do (progn (setq result (interpret-command command))
		       (if (eq result 'quit) (setq live nil))
		       (format t "~%+>")
		       (setq live (not (system-halted system)))))))

(defun ireenter ()
  (cond ((not (boundp '*system*))
	 (format t "No current system (*system*)."))
	((systemp *system*)
	 (irun-system *system*))
	(t (format t "Not systemp - *system*."))))

(defun interpret-command (command)
  (case command
    ((show display) (format t " <what> ") (interpret-show (read)))
    ((set assign)   (format t " <what> ") (interpret-set (read)))
    (event (format t " <event> ") (queue-event (cons 0 (read)) *system*))
    (step  (format t "~%[ ~d] " (system-time *system*))
	   (let ((ev (current-events (system-schedule *system*) (system-time *system*))))
	     (if ev
	       (format t "Events: ~a." ev)
	       (format t "No events occur.  Checking conditions.")))
	   (setq *system* (step-system *system*)))
    ((q quit exit) 'quit)
    ((help ?) (format t "~%Commands: show .., event .., help, step ~%"))
    (otherwise (format t "~%Not a command. Try help."))))

(defun queue-event (ev s)
  (setf (system-schedule s) (cons ev (system-schedule s))))

(defun event-pending (s)
  (let ((sch (system-schedule s)))
    (and (consp sch)
	 (or (equal (caar sch) 0)
	     (equal (caar sch) (system-time s))))))

;; ----------------------------------------------------------------------
;; DISPLAY

(defun interpret-show (command)
  (case command
    (system (display-system *system*))
    ((time configuration alist schedule root nodes arcs prim-events)
     (display-system-component command *system*))
    (node (format t " <name> ") (display-node (find-node (read) (system-nodes *system*))))
    (arc  (format t " <name> ") (display-arc  (find-arc  (read) (system-arcs *system*))))
    (function (format t " <name> ") (display-function  (function-def (read))))
    ((var variable) (format t " <name> ")
                    (display-variable (assoc (read) (system-vars *system*))))
    (otherwise 
     (format t "Not a command. Expecting one of:
 [system 
  time configuration alist schedule root nodes arcs prim-events 
  node arc function var]."))))

(defun interpret-set (command)
  ;; time configuration alist schedule root nodes arcs prim-events)
  (case command
    (time (format t " <integer> ") (setf (system-time *system*) (read)))
    (root (format t " <node name> ") (setf (system-root *system*) (read)))
    (variable (format t " <variable> ")
	      (let ((var (read))
		    (value (progn (format t " <value> ") (read))))
		(setf (system-alist *system*) (bind var value (system-alist *system*)))))
    (otherwise (format t "~%Not settable."))))

(defun display-system (s)		;*system*
  (format t "~%Time: ~d" (system-time s))
  (format t "~%Configuration: ~a" (system-configuration s))
  (format t "~%Alist: ~a" (system-alist s))
  (format t "~%Schedule: ~a" (system-schedule s))
  (format t "~%Root: ~a" (system-root s))
  (format t "~%Nodes: ~a" (mapcar #'node-name (system-nodes s)))
  (format t "~%Arcs: ~a" (mapcar #'arc-name (system-arcs s)))
  (format t "~%Primitive Events: ~a" (system-prim-events s))
  )

(defun display-system-component (c s) ; command *system*
  ;; time configuration alist schedule root nodes arcs prim-events
  (case c
    (time (format t "~%Time: ~d" (system-time s)))
    (configuration (format t "~%Configuration: ~a" (system-configuration s)))
    (alist (format t "~%Alist: ~a" (system-alist s)))
    (schedule (format t "~%Schedule: ~a" (system-schedule s)))
    (root (format t "~%Root: ~a" (system-root s)))
    (nodes (format t "~%Nodes: ~a" (mapcar #'node-name (system-nodes s))))
    (arcs (format t "~%Arcs: ~a" (mapcar #'arc-name (system-arcs s))))
    (prim-events (format t "~%Primitive Events: ~a" (system-prim-events s)))
    (otherwise (format t "~% Not a command. "))))

(defun display-node (n)
  ;; name type level parent children initial enter exit ancestors descendents
  (format t "~%Name  : ~a  - [~a / level ~d / idle ~d / busy: ~d]"
	  (node-name n) (node-type n) (node-level n) (node-idle n) (node-busy n))
  (format t "~%Parent: ~a" (node-parent n))
  (format t "~%Children: ~a" (node-children n))
  (format t "~%Initial : ~a" (node-initial n))
  (format t "~%Enter : ~a" (node-enter n))
  (format t "~%Exit  : ~a" (node-exit n))
  (format t "~%Ancestors  : ~a" (node-ancestors n))
  (format t "~%Descendents: ~a" (node-descendents n))
  )

(defun display-arc (a)
  ;; name from to trigger action scope
  (format t "~%Name  : ~a" (arc-name a))
  (format t "~%From  : ~a" (arc-from a))
  (format t "~%To    : ~a" (arc-to a))
  (format t "~%Trigger : ~a" (arc-trigger a))
  (format t "~%Action  : ~a" (arc-action a))
  (format t "~%Scope   : ~a" (arc-scope a)))

(defun display-enabled-arc (a)
  ;; name from to trigger action scope
  (format t "~%Enables ~a (~a -> ~a) " (arc-name a) (arc-from a) (arc-to a)))

(defun display-function (f)  ; (name . body)
  (format t "~%Name : ~a = ~a" (car f)) (cdr f)))
  
(defun display-variable (v)  ; (name . value)
  (format t "~%Variable : ~a = ~a" (car v) (cdr v)))
 
  
(defun select-by-number (msg list &optional (func #'princ))
  (let ((i 0))
    (format t msg)
    (iterate for x in list do (progn (format t "~%~3d. " i)
				     (apply func (list x))
				     (setq i (+ i 1))))
    (format t "~%> ")
    (let ((n (read)))
      (if (and (numberp n) (< n (length list)))
	  (nth n list)
	  (select-by-number "Response must be a number." list func)))))