;; Parsing input from the MOO

;; <msg>     == ###<request-name>[ <keyword>: <value>]* 
;;  
;; ### must start a line. 
;; End of message indicated by CR.  
;;  

(require "general")
(require "tk")
(provide "mcp-read")

(eval-when (compile)			;;  load eval
 (load-base "conditions.lisp"))

(defvar *mcp-input* t) 

(defvar mcp-readtable (copy-readtable))

(set-syntax-from-char #\# #\a mcp-readtable )
(set-syntax-from-char #\$ #\a mcp-readtable )
(set-syntax-from-char #\, #\  mcp-readtable )

(defvar *mcp-start* "###")
(defvar *mcp-stop-chars*  '(#\newline #\return))

(defvar *mcp-start-len*  (length *mcp-start*))
(defvar *mcp-start-head* (aref *mcp-start* 0))
(defvar *mcp-start-tail* (subseq *mcp-start* 1))

(defvar *mcp-newline* (string #\newline))

(defvar *mcp-read-skip* (list *mcp-start-head* #\newline #\return))

(defvar *pending-mcp-input* nil)

(defun mcp-clear () (setq *pending-mcp-input* nil))

(defun mcp-file-listen () 
  (and (streamp *mcp-input*)
       (not (equal (file-position *mcp-input*)
		   (file-length *mcp-input*)))))

(defun mcp-listen () 
  (or (and *pending-mcp-input* (not (string-equal *pending-mcp-input* "")))
      (mcp-file-listen)))

(defun mcp-read-line ()
  "If line of form `<front>#', with no eol or cr, print <front>, return NIL.
If line of form `<front>###<msg>[\n^M]', print <front>, return <msg>.
If line of form `<front>[\n^M]', print <front>, return T."
  (let ((*readtable* mcp-readtable))
    (multiple-value-bind (line no-end)
	(if *pending-mcp-input*
	    (concatenate 'string *pending-mcp-input* (read-line *mcp-input* nil 'done))
	    (read-line *mcp-input* nil 'done))
      (progn (setq *pending-mcp-input* nil)
	     (cond ((equal line 'done) nil)
		   (no-end (mcp-msg-line line nil))
		   (t (mcp-msg-line line t)))))))

(defvar *mud-log* nil)

(defun mcp-echo (string)
  (if (stringp string) (tk::insert-output string))
  (if (streamp *mud-log*)
      (format *mud-log* (format nil "~a" string))))

(defun mcp-msg-line (line eol)
  (let ((len (length line))
	(i (position *mcp-start-head* line))
	msg)
    (cond ((null i)
	   (mcp-echo line)
	   (if eol
	       (progn (mcp-echo *mcp-newline*) eol)
	       nil))
	  (t (mcp-echo (subseq line 0 i))
	     (cond ((string-equal (subseq line i (min len (+ i *mcp-start-len*))) *mcp-start*)
		    ;;; WIRED IN.
		    (cond ((char-equal (aref line (- len 1)) #\return)
			   (setq msg (subseq line (+ i *mcp-start-len*) (- len 1))))
			  (eol (setq msg (subseq line (+ i *mcp-start-len*))))
			  (t (mcp-unread (subseq line (+ i *mcp-start-len*))))))
		   ((< len *mcp-start-len*)
		    (mcp-unread (mcp-unread (subseq line (+ i *mcp-start-len*)))))
		   (t (mcp-echo (subseq line i))))
	     (if (equal i 0) nil (mcp-echo *mcp-newline*))
	     (or msg eol)))))

(defun mcp-unread (c)
  ;; c may be char or string
  (if (and (stringp c) (string-equal c ""))
      *pending-mcp-input*
      (setq *pending-mcp-input*
	    (concatenate 'string (or *pending-mcp-input* "") (string c)))))

;; Now parse it

;; <msg>     == ###<request-name>[ <keyword>: <value>]* 

;; <char>    == Printing characters (excludes control chars, LF, CR, TAB). 
;; <keyword> == [a-zA-Z]<a>* 
;; <a>       == <char> - { space " \ : } 
;; <value>   == <keyword> | <number> | <obj> | string> 
;; <obj>     == #[0-9]+ 
;; <string>  == "<c>*" 
;; <c>       == <char> - { " } 
;; <number>  == [-][0-9]+ 
;;  
;; I presume this example is legitimate: 
;;  
;; ###describe obj: #23 text: a###b 

(defun mcp-read (msg)
  (if (and (stringp msg) 
	   (> (length msg) 0))
      (let ((*readtable* mcp-readtable))
	(ignore-errors (read-from-string msg)))))

(defun mcp-parse (msg)
  (let ((*readtable* mcp-readtable))
    (multiple-value-bind
	(request i) (ignore-errors (read-from-string msg))
      (if (null request)
	  nil
	  (cons request
		(mcp-parse-keyword-values
		 (substitute #\) #\} (substitute #\( #\{ (subseq msg i)))))))))

(defun mcp-parse-keyword-values (msg)
  (let ((i (position #\: msg)))
    (if (null i)
	nil
	(let ((key (ignore-errors (read-from-string (subseq msg 0 i)))))
	  (multiple-value-bind (value j)
	      (ignore-errors (read-from-string (subseq msg (+ i 1))))
	    (cons (cons key value)
		  (if (numberp j)
		      (mcp-parse-keyword-values (subseq msg (+ j i 1))))))))))

(defun test (file &aux msg)
  (with-open-file (*mcp-input* file :direction :input) 
    (iterate until (null (setq msg (mcp-read-line)))
	     do (if (stringp msg) (format t "---> ~a <----~%" msg)))))
